@@ -9,6 +9,7 @@ use overload
 sub new {
   my $self = bless {}, shift;
   @$self{qw(path message)} = ($_[0] || '/', $_[1] || '');
+  map { $self->{$_} = $_[2]->{$_} } keys %{$_[2]} if $_[2];
   $self;
 }
 
@@ -46,7 +47,7 @@ A human readable description of the error. Defaults to empty string.
 
   $str = $self->path;
 
-A JSON pointer to where the error occured. Defaults to "/".
+A JSON pointer to where the error occurred. Defaults to "/".
 
 =head1 METHODS
 
